from __future__ import annotations

import logging

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


def get_bias(sim: pd.Series[float], obs: pd.Series[float]) -> float:
    return float(sim.mean() - obs.mean())


def get_mse(sim: pd.Series[float], obs: pd.Series[float]) -> float:
    return float(np.square(np.subtract(obs, sim)).mean())


def get_rmse(sim: pd.Series[float], obs: pd.Series[float]) -> float:
    return float(np.sqrt(get_mse(sim, obs)))


def get_mae(sim: pd.Series[float], obs: pd.Series[float]) -> float:
    return float(np.abs(np.subtract(obs, sim)).mean())


def get_mad(sim: pd.Series[float], obs: pd.Series[float]) -> float:
    return float(np.abs(np.subtract(obs, sim)).std())


def get_madp(sim: pd.Series[float], obs: pd.Series[float]) -> float:
    pc1, pc2 = get_percentiles(sim, obs)
    return get_mad(pc1, pc2)


def get_madc(sim: pd.Series[float], obs: pd.Series[float]) -> float:
    madp = get_madp(sim, obs)
    return get_mad(sim, obs) + madp


def get_rms(sim: pd.Series[float], obs: pd.Series[float]) -> float:
    crmsd = ((sim - sim.mean()) - (obs - obs.mean())) ** 2
    return float(np.sqrt(crmsd.mean()))


def get_corr(sim: pd.Series[float], obs: pd.Series[float]) -> float:
    return float(sim.corr(obs))


def get_nse(sim: pd.Series[float], obs: pd.Series[float]) -> float:
    nse = 1 - np.nansum(np.subtract(obs, sim) ** 2) / np.nansum((obs - float(np.nanmean(obs))) ** 2)
    return float(nse)


def get_lambda(sim: pd.Series[float], obs: pd.Series[float]) -> float:
    Xmean = float(np.nanmean(obs))
    Ymean = float(np.nanmean(sim))
    nObs = len(obs)
    corr = get_corr(sim, obs)
    if corr >= 0:
        kappa = 0
    else:
        kappa = 2 * abs(np.nansum((obs - Xmean) * (sim - Ymean)))

    numerator = np.nansum((obs - sim) ** 2)
    denominator = (
        np.nansum((obs - Xmean) ** 2)
        + np.nansum((sim - Ymean) ** 2)
        + nObs * ((Xmean - Ymean) ** 2)
        + kappa
    )
    lambda_index = 1 - numerator / denominator
    return float(lambda_index)


def get_kge(sim: pd.Series[float], obs: pd.Series[float]) -> float:
    corr = get_corr(sim, obs)
    b = (sim.mean() - obs.mean()) / obs.std()
    g = sim.std() / obs.std()
    return float(1 - np.sqrt((corr - 1) ** 2 + b**2 + (g - 1) ** 2))


def truncate_seconds(ts: pd.Series[float]) -> pd.Series[float]:
    df = pd.DataFrame({"time": ts.index, "value": ts.values})
    df = df.assign(time=df.time.dt.floor("min"))
    if df.time.duplicated().any():
        # There are duplicates. Keep the first datapoint per minute.
        msg = "Duplicate timestamps have been detected after the truncation of seconds. Keeping the first datapoint per minute"
        logger.warning(msg)
        df = df.iloc[df.time.drop_duplicates().index].reset_index(drop=True)
    df.index = df.time  # type: ignore[assignment]
    df = df.drop("time", axis=1)
    ts = pd.Series(index=df.index, data=df.value)
    return ts


def align_ts(
    sim: pd.Series[float],
    obs: pd.Series[float],
    resample_to_model: bool = True,
) -> tuple[pd.Series[float], pd.Series[float]]:
    # requisite: obs is the observation time series
    obs = obs.dropna()
    obs = truncate_seconds(obs)
    if resample_to_model:
        freq = sim.index.to_series().diff().median()
        obs = obs.resample(pd.Timedelta(freq)).mean()
    sim_, obs_ = sim.align(obs, axis=0)
    nan_mask1 = pd.isna(sim_)
    nan_mask2 = pd.isna(obs_)
    nan_mask = np.logical_or(nan_mask1, nan_mask2)
    sim_ = sim_[~nan_mask]
    obs_ = obs_[~nan_mask]
    return sim_, obs_


def get_percentiles(
    sim: pd.Series[float],
    obs: pd.Series[float],
    higher_tail: bool = False,
) -> tuple[pd.Series[float], pd.Series[float]]:
    x = np.arange(0, 0.99, 0.01)
    if higher_tail:
        x = np.hstack([x, np.arange(0.99, 1, 0.001)])

    pc_sim = sim.quantile(x).to_numpy()
    pc_obs = obs.quantile(x).to_numpy()

    return pd.Series(pc_sim), pd.Series(pc_obs)


def get_slope_intercept(sim: pd.Series[float], obs: pd.Series[float]) -> tuple[float, float]:
    # Calculate means of x and y
    x_mean = float(np.mean(obs))
    y_mean = float(np.mean(sim))

    numerator = np.sum((obs - x_mean) * (sim - y_mean))
    denominator = np.sum((obs - x_mean) ** 2)

    # Calculate slope (A) and intercept (B) in A*X + B
    slope = numerator / denominator
    intercept = y_mean - slope * x_mean
    return slope, intercept


def get_slope_intercept_pp(sim: pd.Series[float], obs: pd.Series[float]) -> tuple[float, float]:
    pc1, pc2 = get_percentiles(sim, obs)
    slope, intercept = get_slope_intercept(pc1, pc2)
    return slope, intercept



from __future__ import annotations

import logging
from collections.abc import Sequence

import numpy as np
import pandas as pd

from seastats.stats import get_bias
from seastats.stats import get_corr
from seastats.stats import get_kge
from seastats.stats import get_lambda
from seastats.stats import get_mad
from seastats.stats import get_madc
from seastats.stats import get_madp
from seastats.stats import get_mae
from seastats.stats import get_mse
from seastats.stats import get_nse
from seastats.stats import get_rms
from seastats.stats import get_rmse
from seastats.stats import get_slope_intercept
from seastats.stats import get_slope_intercept_pp
from seastats.storms import match_extremes

logger = logging.getLogger(__name__)

GENERAL_METRICS_ALL = [
    "bias",
    "rmse",
    "mae",
    "mse",
    "rms",
    "sim_mean",
    "obs_mean",
    "sim_std",
    "obs_std",
    "nse",
    "lamba",
    "cr",
    "slope",
    "intercept",
    "slope_pp",
    "intercept_pp",
    "mad",
    "madp",
    "madc",
    "kge",
]
GENERAL_METRICS = ["bias", "rms", "rmse", "cr", "nse", "kge"]
STORM_METRICS = ["R1", "R3", "error"]
STORM_METRICS_ALL = ["R1", "R1_norm", "R3", "R3_norm", "error", "error_norm"]

SUGGESTED_METRICS = sorted(GENERAL_METRICS + STORM_METRICS)
SUPPORTED_METRICS = sorted(GENERAL_METRICS_ALL + STORM_METRICS_ALL)


def get_stats(
    sim: pd.Series[float],
    obs: pd.Series[float],
    metrics: Sequence[str] = SUGGESTED_METRICS,
    quantile: float = 0,
    cluster: int = 72,
    round: int = -1,  # noqa: A002
) -> dict[str, float]:
    """
    Calculates various statistical metrics between the simulated and observed time series data.

    :param pd.Series sim: The simulated time series data.
    :param pd.Series obs: The observed time series data.
    :param list[str] metrics: (Optional) The list of statistical metrics to calculate. If metrics = ["all"], all items in SUPPORTED_METRICS will be calculated. Default is all items in SUGGESTED_METRICS.
    :param float quantile: (Optional) Quantile used to calculate the metrics. Default is 0 (no selection)
    :param int cluster: (Optional) Cluster duration for grouping storm events. Default is 72 hours.
    :param int round: (Optional) Apply rounding to the results to. Default is no rounding (value is -1)

    :return dict stats: dictionary containing the calculated metrics.

    The dictionary contains the following keys and their corresponding values:

    - `bias`: The bias between the simulated and observed time series data.
    - `rmse`: The Root Mean Square Error between the simulated and observed time series data.
    - `mae`: The Mean Absolute Error the simulated and observed time series data.
    - `mse`: The Mean Square Error the simulated and observed time series data.
    - `rms`: The raw mean square error between the simulated and observed time series data.
    - `sim_mean`: The mean of the simulated time series data.
    - `obs_mean`: The mean of the observed time series data.
    - `sim_std`: The standard deviation of the simulated time series data.
    - `obs_std`: The standard deviation of the observed time series data.
    - `nse`: The Nash-Sutcliffe efficiency between the simulated and observed time series data.
    - `lamba`: The lambda statistic between the simulated and observed time series data.
    - `cr`: The correlation coefficient between the simulated and observed time series data.
    - `slope`: The slope of the linear regression between the simulated and observed time series data.
    - `intercept`: The intercept of the linear regression between the simulated and observed time series data.
    - `slope_pp`: The slope of the linear regression between the percentiles of the simulated and observed time series data.
    - `intercept_pp`: The intercept of the linear regression between the percentiles of the simulated and observed time series data.
    - `mad`: The median absolute deviation of the simulated time series data from its median.
    - `madp`: The median absolute deviation of the simulated time series data from its median, calculated using the percentiles of the observed time series data.
    - `madc`: The median absolute deviation of the simulated time series data from its median, calculated by adding `mad` to `madp`
    - `kge`: The Kling-Gupta efficiency between the simulated and observed time series data.
    - `R1`: Difference between observed and modelled for the biggest storm
    - `R1_norm`: Normalized R1 (R1 divided by observed value)
    - `R3`: Average difference between observed and modelled for the three biggest storms
    - `R3_norm`: Normalized R3 (R3 divided by observed value)
    - `error`: Average difference between observed and modelled for all storms
    - `error_norm`: Normalized error (error divided by observed value)
    """
    if not isinstance(metrics, list):
        raise ValueError("metrics must be a list")

    if metrics == ["all"]:
        metrics = SUPPORTED_METRICS

    if not np.any([m in SUPPORTED_METRICS for m in metrics]):
        raise ValueError("metrics must be a list of supported variables in SUPPORTED_METRICS or ['all']")

    # Storm metrics part with PoT Selection
    if np.any([m in STORM_METRICS_ALL for m in metrics]):
        extreme_df = match_extremes(sim, obs, quantile=quantile, cluster=cluster)
    else:
        extreme_df = pd.DataFrame()  # Just to make mypy happy

    if quantile:  # signal subsetting is only to be done on general metrics
        sim = sim[sim > sim.quantile(quantile)]
        obs = obs[obs > obs.quantile(quantile)]

    stats = {}

    for metric in metrics:
        match metric:
            case "bias":
                stats["bias"] = get_bias(sim, obs)
            case "rmse":
                stats["rmse"] = get_rmse(sim, obs)
            case "rms":
                stats["rms"] = get_rms(sim, obs)
            case "sim_mean":
                stats["sim_mean"] = sim.mean()
            case "obs_mean":
                stats["obs_mean"] = obs.mean()
            case "sim_std":
                stats["sim_std"] = sim.std()
            case "obs_std":
                stats["obs_std"] = obs.std()
            case "mae":
                stats["mae"] = get_mae(sim, obs)
            case "mse":
                stats["mse"] = get_mse(sim, obs)
            case "nse":
                stats["nse"] = get_nse(sim, obs)
            case "lamba":
                stats["lamba"] = get_lambda(sim, obs)
            case "cr":
                stats["cr"] = get_corr(sim, obs)
            case "slope":
                stats["slope"], _ = get_slope_intercept(sim, obs)
            case "intercept":
                _, stats["intercept"] = get_slope_intercept(sim, obs)
            case "slope_pp":
                stats["slope_pp"], _ = get_slope_intercept_pp(sim, obs)
            case "intercept_pp":
                _, stats["intercept_pp"] = get_slope_intercept_pp(sim, obs)
            case "mad":
                stats["mad"] = get_mad(sim, obs)
            case "madp":
                stats["madp"] = get_madp(sim, obs)
            case "madc":
                stats["madc"] = get_madc(sim, obs)
            case "kge":
                stats["kge"] = get_kge(sim, obs)
            case "R1":
                stats["R1"] = extreme_df["error"].iloc[0]
            case "R1_norm":
                stats["R1_norm"] = extreme_df["error_norm"].iloc[0]
            case "R3":
                stats["R3"] = extreme_df["error"].iloc[0:3].mean()
            case "R3_norm":
                stats["R3_norm"] = extreme_df["error_norm"].iloc[0:3].mean()
            case "error":
                stats["error"] = extreme_df["error"].mean()
            case "error_norm":
                stats["error_norm"] = extreme_df["error_norm"].mean()
    else:
        logger.info("no storm metric specified")

    if round > 0:
        for metric in metrics:
            stats[metric] = np.round(stats[metric], round)

    return stats



from __future__ import annotations

import typing as T

import numpy as np
import pandas as pd
from pyextremes import get_extremes


def match_extremes(
    sim: pd.Series[float],
    obs: pd.Series[float],
    quantile: float,
    cluster: int = 72,
) -> pd.DataFrame:
    """
    Calculate metrics for comparing simulated and observed storm events.
    Parameters:
    - sim (pd.Series): Simulated storm series.
    - obs (pd.Series): Observed storm series.
    - quantile (float): Quantile value for defining extreme events.
    - cluster (int, optional): Cluster duration for grouping storm events. Default is 72 hours.

    Returns:
    - df: pd.DataFrame of the matched extremes between the observed and modeled pd.Series.
          with following columns:
           * `observed`: observed extreme event value
           * `time observed`: observed extreme event time
           * `model`: modeled extreme event value
           * `time model`: modeled extreme event time
           * `diff`: difference between model and observed
           * `error`: absolute difference between model and observed
           * `error_norm`: normalised difference between model and observed
           * `tdiff`: time difference between model and observed (in hours)

    !Important: The modeled values are matched on the observed events calculated by POT analysis.
                The user needs to be mindful about the order of the observed and modeled pd.Series.
    """
    # resample observation to 1H time series
    obs = obs.resample("1h").mean().shift(freq="30min")
    # get observed extremes
    ext = get_extremes(obs, "POT", threshold=obs.quantile(quantile), r=f"{cluster}h")
    ext_values_dict: dict[str, T.Any] = {}
    ext_values_dict["observed"] = ext.values
    ext_values_dict["time observed"] = ext.index.values
    #
    max_in_window = []
    tmax_in_window = []
    # match simulated values with observed events
    for it, itime in enumerate(ext.index):
        snippet = sim[itime - pd.Timedelta(hours=cluster / 2) : itime + pd.Timedelta(hours=cluster / 2)]
        try:
            tmax_in_window.append(snippet.index[int(snippet.argmax())])
            max_in_window.append(snippet.max())
        except Exception:
            tmax_in_window.append(itime)
            max_in_window.append(np.nan)
    ext_values_dict["model"] = max_in_window
    ext_values_dict["time model"] = tmax_in_window
    #
    df = pd.DataFrame(ext_values_dict)
    df = df.dropna(subset="model")
    df = df.sort_values("observed", ascending=False)
    df["diff"] = df["model"] - df["observed"]
    df["error"] = abs(df["diff"])
    df["error_norm"] = abs(df["diff"] / df["observed"])
    df["tdiff"] = df["time model"] - df["time observed"]
    df["tdiff"] = df["tdiff"].apply(lambda x: x.total_seconds() / 3600)
    df = df.set_index("time observed")
    return df
